/*!
 * @file: ws-asm.cpp
 * @brief a compiler to generate whitespace code out of ws assembler
 * @details
 * @author Thomas Geymayer <tomgey@gmail.com>
 * @date Date of Creation: 12.03.2009
 * @version 0.1b
 */

/*
    Copyright (C) 2009 Thomas Geymayer <tomgey@gmail.com>
    TomProgs - http://www.tomprogs.at

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <map>
#include <stack>
#include <boost/regex.hpp>

using std::cout;
using std::endl;

//------------------------------------------------------------------------------

struct Command
{
  std::string param_;
  std::string code_;

  Command(){};

  Command( const char *param, const char *code ):
    param_( param ),
    code_ ( code  )
  {};
};

//------------------------------------------------------------------------------

typedef std::map<std::string,Command>      CommandMap;
typedef std::map<std::string,unsigned int> LabelMap;
typedef std::stack<std::string>            IncludeStack;

std::string convertNumber( unsigned int       number, unsigned int line = 0 );
std::string convertNumber( const std::string& number, unsigned int line = 0 );

//------------------------------------------------------------------------------
int main( int argc, char *argv[] )
{
  if( argc != 2 )
  {
    cout << "usage: ws-asm <input-file>" << endl;
    return 0;
  }

  //----------------------------------------------------------------------------
  // Command list:
  //----------------------------------------------------------------------------

  CommandMap commands;

  commands[ "push"      ] = Command( "number", "  "      );
  commands[ "pushs"     ] = Command( "string", ""        );
  commands[ "pop"       ] = Command( "none",   " \n\n"   );
  commands[ "duplicate" ] = Command( "none",   " \n "    );
  commands[ "swap"      ] = Command( "none",   " \n\t"   );

  commands[ "add"       ] = Command( "pushnum","\t   "   );
  commands[ "sub"       ] = Command( "pushnum","\t  \t"  );
  commands[ "mul"       ] = Command( "pushnum","\t  \n"  );
  commands[ "div"       ] = Command( "pushnum","\t \t "  );
  commands[ "mod"       ] = Command( "pushnum","\t \t\t" );

  commands[ "label"     ] = Command( "nocall", "\n  "    );
  commands[ "call"      ] = Command( "label",  "\n \t"   );
  commands[ "jmp"       ] = Command( "label",  "\n \n"   );
  commands[ "jz"        ] = Command( "label",  "\n\t "   );
  commands[ "jn"        ] = Command( "label",  "\n\t\t"  );
  commands[ "return"    ] = Command( "none",   "\n\t\n"  );
  commands[ "exit"      ] = Command( "none",   "\n\n\n"  );

  commands[ "store"     ] = Command( "pushnsw","\t\t "   );
  commands[ "retrieve"  ] = Command( "pushnum","\t\t\t"  );

  commands[ "putc"      ] = Command( "none",   "\t\n  "  );
  commands[ "putn"      ] = Command( "none",   "\t\n \t" );
  commands[ "getc"      ] = Command( "pushnum","\t\n\t " );
  commands[ "getn"      ] = Command( "pushnum","\t\n\t\t");


  commands[ "include"   ] = Command( "include",""        );

  //----------------------------------------------------------------------------

  cout << "Welcome to the ws-Assembler to Whitespaceconverter." << endl;

  std::ifstream file( argv[1] );

  if( !file )
  {
    std::cerr << "Failed to open input file: " << argv[1] << endl;
    return 1;
  }

  std::string   output;
  LabelMap      labels;

  const boost::regex comment("\\s*#.*$");
  const boost::regex non_ws("^\\s*\\S+.*");
  const boost::regex command("^\\s*(\\l+|\\w+:)"                      // command
                             "(\\s+(([-\\w]+)|\"([^\"]+)\"))?\\s*$"); // param

  IncludeStack includes_;

  do
  {
    if( includes_.size() )
    {
      file.close();
      file.open( includes_.top().c_str() );

      if( !file )
      {
        // maybe its in the lib directory
        includes_.top() = "./lib/" + includes_.top();
        file.open( includes_.top().c_str() );

        if( !file )
        {
          cout << "[ERROR] Failed to parse included file: "
               << includes_.top() << endl;
        }
      }

      cout << "Parsing included file: " << includes_.top() << endl;

      includes_.pop();
    }

    while( !file.eof() )
    {
      static std::string  line;
      static int          line_counter = 0;

      static unsigned int label_index = 0;

      std::getline( file, line );
      ++line_counter;

      // remove empty lines and comments
      line = boost::regex_replace( line, comment, "" );
      if( !line.length() || !boost::regex_match( line, non_ws ) ) continue;

      boost::smatch result;

      if( !boost::regex_match( line, result, command ) )
      {
        cout << "[ERROR] Invalid command at line "
             << line_counter
             << ":\n"
             << line << endl;
        return 1;
      }

      std::string command_name( result.str(1) );

      if( !command_name.length() )
      {
        cout << "[ERROR] No command at line "
             << line_counter
             << ":\n"
             << line << endl;
        return 1;
      }

      std::string encoded_param("");

      // check for a label
      if( *command_name.rbegin() == ':' )
      {
        if( result.str(2).length() || result.str(3).length() )
        {
          cout << "[ERROR] Characters after label at line "
               << line_counter
               << ":\n"
               << line << endl;
          return 1;
        }

        std::string label_name =
          command_name.substr( 0, command_name.length() - 1);

        if( labels.find( label_name ) != labels.end() )
        {
          cout << "[ERROR] Label allready in use, line "
               << line_counter
               << ":\n"
               << label_name << endl;
          return 1;
        }

        command_name  = "label";
        encoded_param = convertNumber( label_index );

        labels[ label_name ] = label_index++;

        //cout << "[LAB]" << label_name << endl;
      }

      CommandMap::iterator command_it = commands.find( command_name );

      if( command_it == commands.end() )
      {
        cout << "[ERROR] Unknown command at line "
             << line_counter
             << ":\n"
             << line << endl;
        return 1;
      }

      if( command_it->second.param_ == "none" )
      {
        if( result.str(2).length() || result.str(3).length() )
        {
          cout << "[ERROR] Command doesn't expect a parameter at line "
               << line_counter
               << ":\n"
               << line << endl;
          return 1;
        }
      }
      else if( command_it->second.param_ == "label" )
      {
        if( !result.str(3).length() )
        {
          cout << "[ERROR] Command needs a target label, line "
               << line_counter
               << ":\n"
               << line << endl;
          return 1;
        }

        encoded_param += result.str(3);
      }
      else if( command_it->second.param_ == "number" )
      {
        if( !result.str(3).length() )
        {
          cout << "[ERROR] Command needs a number, line "
               << line_counter
               << ":\n"
               << line << endl;
          return 1;
        }

        encoded_param += convertNumber( result.str(3) );
      }
      else if( command_it->second.param_ == "pushnum" )
      {
        if( result.str(3).length() )
        {
          CommandMap::iterator it_push = commands.find("push");

          if( it_push == commands.end() )
          {
            cout << "[ERROR] Missing command 'push'. Converter is corrupted."
                 << endl;
            return 1;
          }

          output += it_push->second.code_
                  + convertNumber( result.str(3), line_counter );
        }
      }
      else if( command_it->second.param_ == "pushnsw" )
      {
        if( result.str(3).length() )
        {
          CommandMap::iterator it_push = commands.find("push");
          CommandMap::iterator it_swap = commands.find("swap");

          if( it_push == commands.end() )
          {
            cout << "[ERROR] Missing command 'push'. Converter is corrupted."
                 << endl;
            return 1;
          }
          else if( it_swap == commands.end() )
          {
            cout << "[ERROR] Missing command 'swap'. Converter is corrupted."
                 << endl;
            return 1;
          }

          output += it_push->second.code_
                  + convertNumber( result.str(3), line_counter )
                  + it_swap->second.code_;
        }
      }
      else if( command_it->second.param_ == "string" )
      {
        if( !result.str(3).length() )
        {
          cout << "[ERROR] pushs without string at line "
               << line_counter
               << ":\n"
               << line << endl;
          break;
        }

        CommandMap::iterator it_push = commands.find("push");

        if( it_push == commands.end() )
        {
          cout << "[ERROR] Missing command 'push'. Converter is corrupted."
               << endl;
          return 1;
        }

        output += it_push->second.code_
                + convertNumber( 0, line_counter ); // terminate with null

        std::string str = result.str(3);

        size_t pos_nl = 0;

        while( (pos_nl = str.find("\\n", pos_nl))
               != std::string::npos )
        {
          str.replace( pos_nl, 2, "\n" );
          ++pos_nl;
        }

        // ignore '"' at begin and end
        for( unsigned int i = str.length() - 2; i > 0; --i )
        {
          output += it_push->second.code_
                  + convertNumber( str.at(i), line_counter );
        }

      }
      else if( command_it->second.param_ == "include" )
      {
        if( !result.str(3).length() )
        {
          cout << "[ERROR] Include without filename, line "
               << line_counter
               << ":\n"
               << line << endl;
          return 1;
        }

        if( result.str(3).at(0) == '"' )
        {
          if(     result.str(3).length() < 3
              || *result.str(3).rbegin() != '"' )
          {
            cout << "[ERROR] Invalid filename at line "
                 << line_counter
                 << ":\n"
                 << line << endl;
            return 1;
          }

          includes_.push(
            result.str(3).substr( 1, result.str(3).length() - 2 ) );
        }
        else
        {
          includes_.push( result.str(3)+".wsal" );
        }
      }

      //cout << command_name << encoded_param << endl;

      output += command_it->second.code_ + encoded_param;
    }
  }while( includes_.size() );

  boost::smatch label_matches;

  cout << "Replacing labels..." << endl;

  for( LabelMap::iterator label_it = labels.begin();
       label_it != labels.end();
       ++label_it )
  {
    size_t length = label_it->first.length();
    size_t find   = 0;

    while(    ( find = output.find(label_it->first, find) )
           != std::string::npos )
    {
      if(    output.length() == find + length
          ||    std::string("\t\n ").find( output.at( find + length ) )
             != std::string::npos  )
      {
        output.replace( find, length, convertNumber( label_it->second ) );
        //cout << label_it->first << endl;
      }
      else
      {
        find += length;
      }
    }
  };

  size_t label_start;

  if(    (label_start = output.find_first_not_of("\t\n "))
      != std::string::npos )
  {
    size_t length = output.find_first_of("\t\n ", label_start);

    if( length != std::string::npos ) length -= label_start;

    cout << "[ERROR] Target label doesn't exist: "
         << output.substr( label_start, length ) << endl;
    return 1;
  }

  std::string file_name_out = boost::regex_replace( std::string(argv[1]),
                                                    boost::regex(".wsa$"),
                                                    "" );

  file_name_out += ".ws";

  std::ofstream file_out( file_name_out.c_str() );

  if( !file_out )
  {
    std::cerr << "Failed to open output file: " << file_name_out << endl;
    return 1;
  }

  file_out << output;

  cout << "Successfully written Whitespace-code to " << file_name_out << endl;

  return 0;
}

//------------------------------------------------------------------------------
std::string convertNumber( unsigned int       number, unsigned int line )
{
  std::string result;

  // sign of number
  if( number < 0 )
    result.append("\t");
  else
    result.append(" ");

  unsigned int mask = 1 << (sizeof( unsigned int ) * 8 - 1);
  bool bit_found = false;

  do
  {
    if( number & mask )
    {
      bit_found = true;
      result.append("\t");
    }
    else if( bit_found )
    {
      result.append(" ");
    }
  }while( mask >>= 1 );

  result.append("\n");

  return result;
}

//------------------------------------------------------------------------------
std::string convertNumber( const std::string& number, unsigned int line )
{
  unsigned int val;

  std::istringstream converter( number );

  // check if failed
  if( (converter >> val).fail() )
  {
    cout << "[ERROR] Invalid parameter should be a number, line "
         << line << endl;
    return std::string();
  }

  return convertNumber( val );
}
